"""
.. module:: world
    :platform: Linux, Unix, Windows
    :synopsis: Provides classes for representing a world and its edges.

.. moduleauthor:: Robert Grant <rhgrant10@gmail.com>

"""
        
import json

class World:
    """The nodes and edges of a particular problem.

    Each :class:`World` is created from a list of nodes, a length function, and
    optionally, a name and a description. Additionally, each :class:`World` has
    a UID. The length function must accept nodes as its first two parameters,
    and is responsible for returning the distance between them. It is the 
    responsibility of the :func:`create_edges` to generate the required
    :class:`Edge`\s and initialize them with the correct *length* as returned
    by the length function.
    
    Once created, :class:`World` objects convert the actual nodes into node
    IDs, since solving does not rely on the actual data in the nodes. These are
    accessible via the :attr:`nodes` property. To access the actual nodes,
    simply pass an ID obtained from :attr:`nodes` to the :func:`data` method,
    which will return the node associated with the specified ID.
    
    :class:`Edge`\s are accessible in much the same way, except two node IDs
    must be passed to the :func:`data` method to indicate which nodes start and
    end the :class:`Edge`. For example:
    
    .. code-block:: python
    
        ids = world.nodes
        assert len(ids) > 1
        node0 = world.data(ids[0])
        node1 = world.data(ids[1])
        edge01 = world.data(ids[0], ids[1])
        assert edge01.start == node0
        assert edge01.end == node1
    
    The :func:`reset_pheromone` method provides an easy way to reset the
    pheromone levels of every :class:`Edge` contained in a :class:`World` to a
    given *level*. It should be invoked before attempting to solve a 
    :class:`World` unless a "blank slate" is not desired. Also note that it
    should *not* be called between iterations of the :class:`Solver` because it
    effectively erases the memory of the :class:`Ant` colony solving it.
        
    :param list nodes: a list of nodes
    :param callable lfunc: a function that calculates the distance between
                           two nodes
    :param str name: the name of the world (default is "world#", where
                     "#" is the ``uid`` of the world)
    :param str description: a description of the world (default is None)
    """
    uid = 0

    def __init__(self, nodes, lfunc, **kwargs):
        self.uid = self.__class__.uid
        self.__class__.uid += 1
        self.name = kwargs.get('name', 'world{}'.format(self.uid))
        self.description = kwargs.get('description', None)
        self._nodes = nodes
        self.lfunc = lfunc
        self.edges = self.create_edges()
        
    @property
    def nodes(self):
        """Nodes IDs."""
        return list(range(len(self._nodes)))
    
    def create_edges(self):
        """Create edges from the nodes.
        
        The job of this method is to map node ID pairs to :class:`Edge`
        instances that describe the edge between the nodes at the given
        indices. Note that all of the :class:`Edge`\s are created within this
        method.
        
        :return: a mapping of node ID pairs to :class:`Edge` instances.
        :rtype: :class:`dict`
        """
        edges = {}
        for m in self.nodes:
            for n in self.nodes:
                a, b = self.data(m), self.data(n)
                if a != b:
                    edge = Edge(a, b, length=self.lfunc(a, b))
                    edges[m, n] = edge
        return edges
        
    def reset_pheromone(self, level=0.01):
        """Reset the amount of pheromone on every edge to some base *level*.
        
        Each time a new set of solutions is to be found, the amount of
        pheromone on every edge should be equalized to ensure un-biased initial
        conditions. 
        
        :param float level: amount of pheromone to set on each edge 
                            (default=0.01)
        """
        for edge in self.edges.values():
            edge.pheromone = level
        
    def data(self, idx, idy=None):
        """Return the node data of a single id or the edge data of two ids.

        If only *idx* is specified, return the node with the ID *idx*. If *idy*
        is also specified, return the :class:`Edge` between nodes with indices
        *idx* and *idy*.

        :param int idx: the id of the first node
        :param int idy: the id of the second node (default is None)
        :return: the node with ID *idx* or the :class:`Edge` between nodes
                  with IDs *idx* and *idy*.
        :rtype: node or :class:`Edge`
        """
        try:
            if idy is None:
                return self._nodes[idx]
            else:
                return self.edges[idx, idy]
        except IndexError:
            return None


class Edge:
    """This class represents the link between starting and ending nodes.

    In addition to *start* and *end* nodes, every :class:`Edge` has *length*
    and *pheromone* properties. *length* represents the static, *a priori*
    information, whereas *pheromone* level represents the dynamic, *a
    posteriori* information.
    
    :param Node a: the node at the start of the :class:`Edge`
    :param Node b: the node at the end of the :class:`Edge`
    :param float length: the length of the :class:`Edge` (default=1)
    :param float pheromone: the amount of pheromone on the :class:`Edge` 
                            (default=0.1)
    """
    def __init__(self, a, b, length=None, pheromone=None):
        self.start = a
        self.end = b
        self.length = 1 if length is None else length
        self.pheromone = 0.1 if pheromone is None else pheromone

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return False
        



"""
.. module:: ant
    :platform: Linux, Unix, Windows
    :synopsis: Provides functionality for finding each solution step as well
               as representing a complete solution.

.. moduleauthor:: Robert Grant <rhgrant10@gmail.com>

"""

import sys
import random
import bisect
import itertools
import functools

from .world import World

@functools.total_ordering
class Ant:
    """
    A single independent finder of solutions to a :class:`World`.

    Each :class:`Ant` finds a solution to a world one move at a time.  They
    also represent the solution they find, and are capable of reporting which
    nodes and edges they visited, in what order they were visited, and the
    total length of the solution.

    Two properties govern the decisions each :class:`Ant` makes while finding
    a solution: *alpha* and *beta*. *alpha* controls the importance placed on
    pheromone while *beta* controls the importance placed on distance. In 
    general, *beta* should be greater than *alpha* for best results.
    :class:`Ant`\s also have a *uid* property that can be used to identify a
    particular instance.

    Using the :func:`initialize` method, each :class:`Ant` *must be 
    initialized* to a particular :class:`World`, and optionally may be given an
    initial node from which to start finding a solution. If a starting node is
    not given, one is chosen at random. Thus a few examples of instantiation
    and initialization might look like:

    .. code-block:: python

        ant = Ant()
        ant.initialize(world)
        
    .. code-block:: python

        ant = Ant().initialize(world)

    .. code-block:: python

        ant = Ant(alpha=0.5, beta=2.25)
        ant.initialize(world, start=world.nodes[0])

    .. note::

        The examples above assume the world has already been created!

    Once an :class:`Ant` has found a solution (or at any time), the solution
    may be obtained and inspected by accessing its ``tour`` property, which
    returns the nodes visited in order, or it's ``path`` property, which 
    returns the edges visited in order. Also, the total distance of the 
    solution can be accessed through its ``distance`` property. :class:`Ant`\s
    are even sortable by their distance:
    
    .. code-block:: python
    
        ants = [Ant() for ...]
        # ... have each ant in the list solve a world
        ants = sorted(ants)
        for i in range(1, len(ants)):
            assert ants[i - 1].distance < ants[i].distance
            
    :class:`Ant`\s may be cloned, which will return a shallow copy while not 
    preserving the *uid* property. If this behavior is not desired, simply use
    the :func:`copy.copy` or :func:`copy.deepcopy` methods as necessary.
    
    The remaining methods mainly govern the mechanics of making each move.
    :func:`can_move` determines whether all possible moves have been made, 
    :func:`remaining_moves` returns the moves not yet made, :func:`choose_move`
    returns a single move from a list of moves, :func:`make_move` actually
    performs the move, and :func:`weigh` returns the weight of a given move.
    The :func:`move` method governs the move-making process by gathering the
    remaining moves, choosing one of them, making the chosen move, and 
    returning the move that was made.
    """
    uid = 0

    def __init__(self, alpha=1, beta=3):
        """Create a new Ant for the given world.

        :param float alpha: the relative importance of pheromone (default=1)
        :param float beta: the relative importance of distance (default=3)
        """
        self.uid = self.__class__.uid
        self.__class__.uid += 1
        self.world = None
        self.alpha = alpha
        self.beta = beta
        self.start = None
        self.distance = 0
        self.visited = []
        self.unvisited = []
        self.traveled = []

    def initialize(self, world, start=None):
        """Reset everything so that a new solution can be found.

        :param World world: the world to solve
        :param Node start: the starting node (default is chosen randomly)
        :return: `self`
        :rtype: :class:`Ant`
        """
        self.world = world
        if start is None:
            self.start = random.randrange(len(self.world.nodes))
        else:
            self.start = start
        self.distance = 0
        self.visited = [self.start]
        self.unvisited = [n for n in self.world.nodes if n != self.start]
        self.traveled = []
        return self

    def clone(self):
        """Return a shallow copy with a new UID.
        
        If an exact copy (including the uid) is desired, use the 
        :func:`copy.copy` method.

        :return: a clone
        :rtype: :class:`Ant`
        """
        ant = Ant(self.alpha, self.beta)
        ant.world = self.world
        ant.start = self.start
        ant.visited = self.visited[:]
        ant.unvisited = self.unvisited[:]
        ant.traveled = self.traveled[:]
        ant.distance = self.distance
        return ant

    @property
    def node(self):
        """Most recently visited node."""
        try:
            return self.visited[-1]
        except IndexError:
            return None

    @property
    def tour(self):
        """Nodes visited by the :class:`Ant` in order."""
        return [self.world.data(i) for i in self.visited]

    @property
    def path(self):
        """Edges traveled by the :class:`Ant` in order."""
        return [edge for edge in self.traveled]
        
    def __eq__(self, other):
        """Return ``True`` if the distance is equal to the other distance.
        
        :param Ant other: right-hand argument
        :rtype: bool
        """
        return self.distance == other.distance

    def __lt__(self, other):
        """Return ``True`` if the distance is less than the other distance.
        
        :param Ant other: right-hand argument
        :rtype: bool
        """
        return self.distance < other.distance

    def can_move(self):
        """Return ``True`` if there are moves that have not yet been made.
    
        :rtype: bool
        """
        # This is only true after we have made the move back to the starting
        # node.
        return len(self.traveled) != len(self.visited)

    def move(self):
        """Choose, make, and return a move from the remaining moves.
        
        :return: the :class:`Edge` taken to make the move chosen
        :rtype: :class:`Edge`
        """
        remaining = self.remaining_moves()
        choice = self.choose_move(remaining)
        return self.make_move(choice)

    def remaining_moves(self):
        """Return the moves that remain to be made.
        
        :rtype: list
        """
        return self.unvisited

    def choose_move(self, choices):
        """Choose a move from all possible moves.
        
        :param list choices: a list of all possible moves
        :return: the chosen element from *choices*
        :rtype: node
        """
        if len(choices) == 0:
            return None
        if len(choices) == 1:
            return choices[0]
        
        # Find the weight of the edges that take us to each of the choices.
        weights = []
        for move in choices:
            edge = self.world.edges[self.node, move]
            weights.append(self.weigh(edge))
        
        # Choose one of them using a weighted probability.
        total = sum(weights)
        cumdist = list(itertools.accumulate(weights)) + [total]
        return choices[bisect.bisect(cumdist, random.random() * total)]

    def make_move(self, dest):
        """Move to the *dest* node and return the edge traveled.
        
        When *dest* is ``None``, an attempt to take the final move back to the
        starting node is made. If that is not possible (because it has 
        previously been done), then ``None`` is returned.
        
        :param node dest: the destination node for the move
        :return: the edge taken to get to *dest*
        :rtype: :class:`Edge`
        """
        # Since self.node simply refers to self.visited[-1], which will be
        # changed before we return to calling code, store a reference now.
        ori = self.node

        # When dest is None, all nodes have been visited but we may not
        # have returned to the node on which we started. If we have, then
        # just do nothing and return None. Otherwise, set the dest to the
        # node on which we started and don't try to move it from unvisited
        # to visited because it was the first one to be moved.
        if dest is None:
            if self.can_move() is False:
                return None
            dest = self.start   # last move is back to the start
        else:
            self.visited.append(dest)
            self.unvisited.remove(dest)
        
        edge = self.world.edges[ori, dest]
        self.traveled.append(edge)
        self.distance += edge.length
        return edge

    def weigh(self, edge):
        """Calculate the weight of the given *edge*.
        
        The weight of an edge is simply a representation of its perceived value
        in finding a shorter solution. Larger weights increase the odds of the
        edge being taken, whereas smaller weights decrease those odds.
        
        :param Edge edge: the edge to weigh
        :return: the weight of *edge*
        :rtype: float
        """
        pre = 1 / (edge.length or 1)
        post = edge.pheromone
        return post ** self.alpha * pre ** self.beta

    



from ..world import World, Edge
import unittest
import math

class WorldTest(unittest.TestCase):
    def setUp(self):
        self.nodeA = dict(name="A")
        self.nodeB = dict(name="B")
        self.nodeC = dict(name="C")
        self.edgeAB = Edge(self.nodeA, self.nodeB)
        self.edgeBA = Edge(self.nodeB, self.nodeA)
        self.edgeAC = Edge(self.nodeA, self.nodeC)
        self.edgeCA = Edge(self.nodeC, self.nodeA)
        self.edgeBC = Edge(self.nodeB, self.nodeC)
        self.edgeCB = Edge(self.nodeC, self.nodeB)
        self.edgeAB2 = Edge(self.nodeA, self.nodeB,
            length=self.edgeAB.length + 1)

class WorldCreation(WorldTest):
    def test_world_creation_with_no_arguments(self):
        w = World()
        self.assertEqual(w.nodes, [])
        self.assertEqual(w.edges, {})
        
    def test_world_creation_from_edge_list(self):
        edges = [self.edgeAB, self.edgeBA]
        w = World(edges)
        self.assertEqual(len(w.nodes), 2)

    def test_world_creation_failure_from_list_of_nonedges(self):
        not_edges = {"Not", "an", "edge", "list"}
        self.assertRaises(TypeError, World, (not_edges,))
    
    def test_world_creation_failure_from_extra_kwargs(self):
        self.assertRaises(TypeError, World, unknown="keyword")
        
    def test_world_creation_failure_from_extra_args(self):
        self.assertRaises(TypeError, World, (1,2,3))
    
    
class EdgeAddition(WorldTest):
    def test_world_correct_edge_count(self):
        w = World()
        w.add_edge(self.edgeAB)
        w.add_edge(self.edgeBA)
        self.assertEqual(len(w.edges), 2, "Wrong number of edges added")

    def test_world_no_duplicate_nodes(self):
        w = World()
        w.add_edge(self.edgeAB)
        w.add_edge(self.edgeAC)
        self.assertEqual(len(w.nodes), 3, "Wrong number of nodes extracted")
        
    def test_world_edge_replacement(self):
        w = World()
        w.add_edge(self.edgeAB)
        w.add_edge(self.edgeAB2)
        self.assertEqual(len(w.edges), 1)
        self.assertEqual(w.edges[0, 1], self.edgeAB2)
        

class PheromoneReset(WorldTest):
    def test_world_pheromone_level_of_every_edge_reset(self):
        edges = [self.edgeAB, self.edgeBA]
        w = World(edges)
        p = 1234
        w.reset_pheromone(p)
        self.assertTrue(all(e.pheromone == p for e in w.edges.values()))
        
    def test_world_pheromone_can_never_be_zero(self):
        edges = [self.edgeAB, self.edgeBA]
        w = World(edges)
        p = 0
        w.reset_pheromone(p)
        self.assertTrue(all(e.pheromone > 0 for e in w.edges.values()))
        

class EdgeTest(unittest.TestCase):
    def setUp(self):
        self.nodeA = dict(name="A")
        self.nodeB = dict(name="B")
        self.nodeC = dict(name="C")


class EdgeCreation(EdgeTest):        
    def test_edge_creation_with_no_args(self):
        self.assertRaises(TypeError, Edge)
        
    def test_edge_creation_with_mutable_nodes(self):
        self.assertRaises(TypeError, Edge, ({}, {}))
        
    def test_edge_creation_with_immutable_nodes(self):
        e = Edge(self.nodeA, self.nodeB)
        self.assertEqual(e.start, self.nodeA)
        self.assertEqual(e.end, self.nodeB)
        
    def test_edge_creation_with_custom_length(self):
        length = 100
        e = Edge(self.nodeA, self.nodeB, length=length)
        self.assertEqual(e.length, length)
        
    def test_edge_creation_with_custom_pheromone_level(self):
        level = 100
        e = Edge(self.nodeA, self.nodeB, pheromone=level)
        self.assertEqual(e.pheromone, level)


class EdgeEquality(EdgeTest):
    def test_edge_identical_edge_equality(self):
        e1 = Edge(self.nodeA, self.nodeB)
        e2 = Edge(self.nodeA, self.nodeB)
        self.assertEqual(e1, e2)
        
    def test_edge_length_difference_inequality(self):
        e1 = Edge(self.nodeA, self.nodeB, length=1)
        e2 = Edge(self.nodeA, self.nodeB, length=2)
        self.assertNotEqual(e1, e2)
        
    def test_edge_pheromone_difference_inequality(self):
        e1 = Edge(self.nodeA, self.nodeB, pheromone=1)
        e2 = Edge(self.nodeA, self.nodeB, pheromone=2)
        self.assertNotEqual(e1, e2)
        
    def test_edge_node_difference_inequality(self):
        e1 = Edge(self.nodeA, self.nodeB)
        e2 = Edge(self.nodeB, self.nodeA)
        self.assertNotEqual(e1, e2)
            
            
if __name__ == '__main__':
    unittest.main()


"""A Python3 implementation of the Ant Colony Optimization Meta-Heuristic.

**Pants** provides you with the ability to quickly determine how to
visit a collection of interconnected nodes such that the work done is
minimized. Nodes can be any arbitrary collection of data while the edges
represent the amount of "work" required to travel between two nodes.
Thus, **Pants** is a tool for solving traveling salesman problems.

The world is built from a list of nodes and a function responsible for
returning the length of the edge between any two given nodes. The length
function need not return actual length. Instead, "length" refers to that 
the amount of "work" involved in moving from the first node to the second
node - whatever that "work" may be. For a silly, random example, it could
even be the number of dishes one must wash before moving to the next 
station at a least dish-washing dish washer competition.

Solutions are found through an iterative process. In each iteration,
several ants are allowed to find a solution that "visits" every node of
the world. The amount of pheromone on each edge is updated according to
its the length of solutions in which it was used. The ant that traveled the
least distance is considered to be the local best solution. If the local
solution has a shorter distance than the best from any previous
iteration, it then becomes the global best solution. The elite ant(s)
then deposit their pheromone along the path of the global best solution
to strengthen it further, and the process repeats.

You can read more about `Ant Colony Optimization on
Wikipedia <http://en.wikipedia.org/wiki/Ant_colony_optimization_algorithms>`_.

.. moduleauthor:: Robert Grant <rhgrant10@gmail.com>

"""

from .ant import Ant 
from .world import World, Edge
from .solver import Solver



"""
.. module:: solver
    :platform: Linux, Unix, Windows
    :synopsis: Provides functionality for finding a complete solution to a 
               world.

.. moduleauthor:: Robert Grant <rhgrant10@gmail.com>

"""

import random
from copy import copy

from .world import World
from .ant import Ant

class Solver:
    """This class contains the functionality for finding one or more solutions
    for a given :class:`World`.
    
    :param float alpha: relative importance of pheromone (default=1)
    :param float beta: relative importance of distance (default=3)
    :param float rho: percent evaporation of pheromone (0..1, default=0.8)
    :param float q: total pheromone deposited by each :class:`Ant` after
                    each iteration is complete (>0, default=1)
    :param float t0: initial pheromone level along each :class:`Edge` of the
                     :class:`World` (>0, default=0.01)
    :param int limit: number of iterations to perform (default=100)
    :param float ant_count: how many :class:`Ant`\s will be used 
                            (default=10)
    :param float elite: multiplier of the pheromone deposited by the elite
                        :class:`Ant` (default=0.5)
    """
    def __init__(self, **kwargs):
        self.alpha = kwargs.get('alpha', 1)
        self.beta = kwargs.get('beta', 3)
        self.rho = kwargs.get('rho', 0.8)
        self.q = kwargs.get('Q', 1)
        self.t0 = kwargs.get('t0', .01)
        self.limit = kwargs.get('limit', 100)
        self.ant_count = kwargs.get('ant_count', 10)
        self.elite = kwargs.get('elite', .5)
        
    def create_colony(self, world):
        """Create a set of :class:`Ant`\s and initialize them to the given 
        *world*.
        
        If the *ant_count* is less than `1`, :func:`round_robin_ants` are
        used and the number of :class:`Ant`\s will be equal to the number of
        nodes. Otherwise, :func:`random_ants` are created instead, and the 
        number of :class:`Ant`\s will be equal to the *ant_count*.
        
        :param World world: the world from which the :class:`Ant`\s will be
                            given starting nodes.
        :return: list of :class:`Ant`\s
        :rtype: list
        """
        if self.ant_count < 1:
            return self.round_robin_ants(world, len(world.nodes))
        return self.random_ants(world, self.ant_count)
        
    def reset_colony(self, colony):
        """Reset the *colony* of :class:`Ant`\s such that each :class:`Ant` is
        ready to find a new solution.
        
        Essentially, this method re-initializes all :class:`Ant`\s in the
        colony to the :class:`World` that they were initialized to last.
        Internally, this method is called after each iteration of the
        :class:`Solver`.
        
        :param list colony: the :class:`Ant`\s to reset
        """
        for ant in colony:
            ant.initialize(ant.world)
        
    def aco(self, colony):
        """Return the best solution by performing the ACO meta-heuristic.
        
        This method lets every :class:`Ant` in the colony find a solution,
        updates the pheromone levels according to the solutions found, and
        returns the `Ant` with the best solution.
        
        This method is not meant to be called directly. Instead, call either
        :func:`solve` or :func:`solutions`.
        
        :param list colony: the `Ant`\s to use in finding a solution
        :return: the best solution found
        :rtype: :class:`Ant`
        """
        self.find_solutions(colony)
        self.global_update(colony)
        return sorted(colony)[0]
        
    def solve(self, world):
        """Return the single shortest path found through the given *world*.

        :param World world: the :class:`World` to solve
        :return: the single best solution found
        :rtype: :class:`Ant`
        """
        world.reset_pheromone(self.t0)
        global_best = None
        colony = self.create_colony(world)
        for i in range(self.limit):
            self.reset_colony(colony)
            local_best = self.aco(colony)
            if global_best is None or local_best < global_best:
                global_best = copy(local_best)
            self.trace_elite(global_best)
        return global_best
    
    def solutions(self, world):
        """Return successively shorter paths through the given *world*.

        Unlike :func:`solve`, this method returns one solution for each 
        improvement of the best solution found thus far. 

        :param World world: the :class:`World` to solve
        :return: successively shorter solutions as :class:`Ant`\s
        :rtype: list
        """
        world.reset_pheromone(self.t0)
        global_best = None
        colony = self.create_colony(world)
        for i in range(self.limit):
            self.reset_colony(colony)
            local_best = self.aco(colony)
            if global_best is None or local_best < global_best:
                global_best = copy(local_best)
                yield global_best
            self.trace_elite(global_best)
    
    def round_robin_ants(self, world, count):
        """Returns a list of :class:`Ant`\s distributed to the nodes of the 
        world in a round-robin fashion.

        Note that this does not ensure at least one :class:`Ant` begins at each
        node unless there are exactly as many :class:`Ant`\s as there are
        nodes. However, if *ant_count* is ``0`` then *ant_count* is set to the
        number of nodes in the :class:`World` and this method is used to create
        the :class:`Ant`\s before solving.

        :param World world: the :class:`World` in which to create the
                            :class:`Ant`\s
        :param int count: the number of :class:`Ant`\s to create
        :return: the :class:`Ant`\s initialized to nodes in the :class:`World`
        :rtype: list
        """
        starts = world.nodes
        n = len(starts)
        return [
            Ant(self.alpha, self.beta).initialize(
                world, start=starts[i % n])
            for i in range(count)
        ]
        
    def random_ants(self, world, count, even=False):
        """Returns a list of :class:`Ant`\s distributed to the nodes of the 
        world in a random fashion.

        Note that this does not ensure at least one :class:`Ant` begins at each
        node unless there are exactly as many :class:`Ant`\s as there are
        nodes. This method is used to create the :class:`Ant`\s before solving
        if *ant_count* is **not** ``0``.

        :param World world: the :class:`World` in which to create the ants.
        :param int count: the number of :class:`Ant`\s to create
        :param bool even: ``True`` if :func:`random.random` should avoid 
                          choosing the same starting node multiple times
                          (default is ``False``)
        :return: the :class:`Ant`\s initialized to nodes in the :class:`World`
        :rtype: list
        """
        ants = []
        starts = world.nodes
        n = len(starts)
        if even:
            # Since the caller wants an even distribution, use a round-robin 
            # method until the number of ants left to create is less than the
            # number of nodes.
            if count > n:
                for i in range(self.ant_count // n):
                    ants.extend([
                        Ant(self.alpha,self.beta).initialize(
                            world, start=starts[j])
                        for j in range(n)
                    ])
            # Now (without choosing the same node twice) choose the reamining
            # starts randomly.
            ants.extend([
                Ant(self.alpha, self.beta).initialize(
                    world, start=starts.pop(random.randrange(n - i)))
                for i in range(count % n)
            ])
        else:
            # Just pick random nodes.
            ants.extend([
                Ant(self.alpha, self.beta).initialize(
                    world, start=starts[random.randrange(n)]) 
                for i in range(count)
            ])
        return ants

    def find_solutions(self, ants):
        """Let each :class:`Ant` find a solution.

        Makes each :class:`Ant` move until each can no longer move.

        .. todo:: 
        
            Make the local pheromone update optional and configurable.

        :param list ants: the ants to use for solving
        """
        # This loop occurs exactly as many times as there are ants times nodes,
        # but that is only because every ant must visit every node. It may be
        # more efficient to convert it to a counting loop... but what 
        # flexibility would we loose in terms of extending the solver class?
        ants_done = 0
        while ants_done < len(ants):
            ants_done = 0
            for ant in ants:
                if ant.can_move():
                    edge = ant.move()
                    self.local_update(edge)
                else:
                    ants_done += 1

    def local_update(self, edge):
        """Evaporate some of the pheromone on the given *edge*.
        
        .. note:: 
        
            This method should never let the pheromone on an edge decrease to 
            less than its initial level.

        :param Edge edge: the :class:`Edge` to be updated
        """
        edge.pheromone = max(self.t0, edge.pheromone * self.rho)

    def global_update(self, ants):
        """Update the amount of pheromone on each edge according to the fitness
        of solutions that use it.

        This accomplishes the global update performed at the end of each
        solving iteration. 
        
        .. note:: 
        
            This method should never let the pheromone on an edge decrease to 
            less than its initial level.

        :param list ants: the ants to use for solving
        """
        ants = sorted(ants)[:len(ants) // 2]
        for a in ants:
            p = self.q / a.distance
            for edge in a.path:
                edge.pheromone = max(
                    self.t0,
                    (1 - self.rho) * edge.pheromone + p)

    def trace_elite(self, ant):
        """Deposit pheromone along the path of a particular ant.

        This method is used to deposit the pheromone of the elite :class:`Ant`
        at the end of each iteration.
        
        .. note:: 
        
            This method should never let the pheromone on an edge decrease to 
            less than its initial level.

        :param Ant ant: the elite :class:`Ant`
        """
        if self.elite:
            p = self.elite * self.q / ant.distance
            for edge in ant.path:
                edge.pheromone += p
    



from ..ant import Ant
from ..world import World, Edge, Node
from ..solver import Solver

import unittest
from unittest import mock

class SolverTest(unittest.TestCase):
    def setUp(self):
        self.nodes = [Node(name='a'), Node(name='b')]
        self.world = mock.Mock(World)
        self.world.nodes = self.nodes[:]
        
    def test_random_ants_when_even_and_ant_count_less_than_node_count(self):
        freq = 2
        node_count = len(self.nodes)
        ant_count = node_count // freq
        solver = Solver(self.world, ant_count=ant_count)
        
        ants = solver.random_ants(even=True)
        
        ants_on = dict.fromkeys(self.nodes, 0)
        for a in ants:
            self.assertFalse(a.start is None,
                    "At least one ant was not placed on a node to start"
                    )
            ants_on[a.start] += 1
        
        self.assertFalse(any(c > 1 for c in ants_on.values()),
                "More than one ant was placed on the same node to start"
                )
        
    def test_random_ants_when_even_and_ant_count_exceeds_node_count(self):
        freq = 2
        node_count = len(self.nodes)
        ant_count = freq * node_count
        solver = Solver(self.world, ant_count=ant_count)
        
        ants = solver.random_ants(even=True)
        
        ants_on = dict.fromkeys(self.nodes, 0)
        for a in ants:
            self.assertFalse(a.start is None,
                    "At least one ant was not placed on a node to start"
                    )
            ants_on[a.start] += 1
        
        self.assertTrue(all(c == freq for c in ants_on.values()),
                "Ants were not placed on nodes to start evenly")
        self.assertEqual(sum(ants_on.values()), ant_count,
                "Not all ants were placed on a node to start")
        
        
if __name__ == '__main__':
    unittest.main()


from ..ant import Ant
from ..world import World, Edge
import unittest
import unittest.mock

class AntTest(unittest.TestCase):
    @unittest.mock.patch('__main__.World')
    def setUp(self, MockWorld):
        self.nodeA = dict(name='a')
        self.nodeB = dict(name='b')
        self.world = World()
        self.aa = 1
        self.ab = 3
        self.ba = 5
        self.edgeAA = Edge(self.nodeA, self.nodeA, length=self.aa)
        self.edgeAB = Edge(self.nodeA, self.nodeB, length=self.ab)
        self.edgeBA = Edge(self.nodeB, self.nodeA, length=self.ba)
        self.edgeNoLength = Edge(self.nodeA, self.nodeB, length=0)
        self.edgeNoPheromone = Edge(self.nodeA, self.nodeB, length=1, pheromone=0)
        self.edge1and1 = Edge(self.nodeA, self.nodeB, 1, 1)

    def test_ant_uids_increase_by_one(self):
        n = 10
        uids = [Ant(self.world).uid for _ in range(n)]
        for i in range(n - 1):
            self.assertTrue(uids[i] + 1 == uids[i + 1])
    
    def test_ant_make_move_adds_no_distance_when_moving_to_start_node(self):
        self.world.add_edge(self.edgeAB)
        ant = Ant(self.world)
        ant.make_move(0)
        self.assertEqual(ant.distance, 0)

    def test_ant_make_move_accounts_for_distance_to_start_node(self):
        def get_edge(pair):
            d = {
                (0, 1): self.edgeAB,
                (1, 0): self.edgeBA
            }
            return d.get(pair, None)
        self.world.edges.__getitem__.side_effect = get_edge
        self.world.nodes = [0, 1]
        ant = Ant(self.world, start=0)
        ant.make_move(1)
        self.assertEqual(ant.distance, self.ab + self.ba)
                
    def test_ant_possible_moves_is_b_when_b_is_only_move_left(self):
        self.world.nodes = [0, 1]
        self.world.edges.__getitem__.side_effect = lambda x: self.edgeAB
        ant = Ant(self.world, start=0)
        moves = list(ant.get_possible_moves())
        self.assertEqual(len(moves), 1)
        self.assertEqual(moves[0], 1)
        
    def test_ant_possible_moves_is_empty_when_no_moves_left(self):
        self.world.nodes = [0]
        self.world.edges.__getitem__.side_effect = lambda x: self.edgeAA
        ant = Ant(self.world, start=0)
        moves = list(ant.get_possible_moves())
        self.assertEqual(len(moves), 0)
    
    def test_ant_choose_move_returns_b_when_b_is_only_move(self):
        self.world.edges.__getitem__.return_value = self.edgeAB
        ant = Ant(self.world, start=0)
        move = ant.choose_move([1])
        self.assertEqual(move, 1)
    
    def test_ant_choose_move_returns_None_when_no_moves(self):
        self.world.add_edge(self.edgeAB)
        ant = Ant(self.world, start=0)
        move = ant.choose_move([])
        self.assertEqual(move, None)

    def test_ant_calculate_weight_when_edge_length_is_zero(self):
        ant = Ant(self.world, alpha=1, beta=1)
        weight = ant.calculate_weight(self.edgeNoLength)
        self.assertEqual(self.edgeNoLength.pheromone, weight)
        
    def test_ant_calculate_weight_when_edge_pheromone_is_zero(self):
        ant = Ant(self.world, alpha=1, beta=1)
        weight = ant.calculate_weight(self.edgeNoPheromone)
        self.assertEqual(weight, 0)
        
    def test_ant_clone_has_equal_properties_after_moving(self):
        self.world.add_edge(self.edgeAB)
        self.world.add_edge(self.edgeBA)
        ant = Ant(self.world)
        ant.move()
        clone = ant.clone()
        self.assertEqual(ant.alpha, clone.alpha)
        self.assertEqual(ant.beta, clone.beta)
        self.assertEqual(ant.start, clone.start)
        self.assertEqual(ant.node, clone.node)
        self.assertEqual(ant.distance, clone.distance)
        self.assertEqual(ant.path, clone.path)
        self.assertEqual(ant.uid, clone.uid)

    def test_ant_clone_has_equal_properties_before_moving(self):
        ant = Ant(self.world)
        clone = ant.clone()
        self.assertEqual(ant.alpha, clone.alpha)
        self.assertEqual(ant.beta, clone.beta)
        self.assertEqual(ant.start, clone.start)
        self.assertEqual(ant.node, clone.node)
        self.assertEqual(ant.distance, clone.distance)
        self.assertEqual(ant.path, clone.path)
        self.assertEqual(ant.uid, clone.uid)
        

if __name__ == '__main__':
    unittest.main()


from distutils.core import setup
import sys

if sys.version_info.major < 3:
    print("Pants requires Python version >= 3, please upgrade")
    sys.exit(1)

setup(
    name="ACO-Pants",
    version="0.5.1",
    author="Robert Grant",
    author_email="rhgrant10@gmail.com",
    packages=["pants", "pants.test"],
    scripts=["bin/pants-demo"],
    url="http://pypi.python.org/pypi/ACO-Pants",
    license="LICENSE.txt",
    description="A Python3 implementation of the ACO Meta-Heuristic",
    long_description=open("README.rst").read()
)